@@ -2,7 +2,7 @@
  * Orthanc - A Lightweight, RESTful DICOM Store
  * Copyright (C) 2012-2016 Sebastien Jodogne, Medical Physics
  * Department, University Hospital of Liege, Belgium
- * Copyright (C) 2017-2018 Osimis S.A., Belgium
+ * Copyright (C) 2017-2019 Osimis S.A., Belgium
  *
  * This program is free software: you can redistribute it and/or
  * modify it under the terms of the GNU Affero General Public License
@@ -24,6 +24,9 @@
 #include "../Common/ImplicitTransaction.h"
 
 #include <Core/Logging.h>
+#include <Core/Toolbox.h>
+#include <Plugins/Samples/Common/OrthancPluginCppWrapper.h>
+
 
 namespace OrthancDatabases
 {
@@ -43,6 +46,7 @@ namespace OrthancDatabases
                         bool isIndex)
   {
     Orthanc::Logging::Initialize(context);
+    OrthancPlugins::SetGlobalContext(context);
     ImplicitTransaction::SetErrorOnDoubleExecution(false);
 
     assert(DisplayPerformanceWarning(dbms, isIndex));
@@ -61,7 +65,7 @@ namespace OrthancDatabases
       return false;
     }
 
-    if (OrthancPluginCheckVersionAdvanced(context, 1, 4, 0) == 1)
+    if (OrthancPluginCheckVersionAdvanced(context, 1, 5, 4) == 1)
     {
       ImplicitTransaction::SetErrorOnDoubleExecution(true);
       isOptimal = true;
@@ -83,12 +87,41 @@ namespace OrthancDatabases
       return false;
     }
 
+    if (useFallback)
+    {
+      std::string v(context->orthancVersion);
+
+      if (v == "mainline")
+      {
+        isOptimal = true;
+      }
+      else
+      {
+        std::vector<std::string> tokens;
+        Orthanc::Toolbox::TokenizeString(tokens, v, '.');
+        
+        if (tokens.size() != 3)
+        {
+          LOG(ERROR) << "Bad version of Orthanc: " << v;
+          return false;
+        }
+
+        int major = boost::lexical_cast<int>(tokens[0]);
+        int minor = boost::lexical_cast<int>(tokens[1]);
+        int revision = boost::lexical_cast<int>(tokens[2]);
+
+        isOptimal = (major > 1 ||
+                     (major == 1 && minor > 5) ||
+                     (major == 1 && minor == 5 && revision >= 4));
+      }
+    }
+
     if (!isOptimal &&
         isIndex)
     {
       LOG(WARNING) << "Performance warning in " << dbms
                    << " index: Your version of Orthanc (" 
-                   << context->orthancVersion << ") should be upgraded to 1.4.0 "
+                   << context->orthancVersion << ") should be upgraded to 1.5.4 "
                    << "to benefit from best performance";
     }
 
